<?php

namespace Drupal\Tests\inline_translation\Functional;

// phpcs:disable Drupal.Arrays.Array.LongLineDeclaration
use Drupal\Tests\content_translation\Functional\ContentTranslationUITestBase;
use Drupal\Tests\file\Functional\FileFieldCreationTrait;
use Drupal\Tests\image\Kernel\ImageFieldCreationTrait;
use Drupal\Tests\TestFileCreationTrait;

/**
 * Tests the Inline Translation UI.
 *
 * @group inline_translation
 */
class InlineTranslationNodeUITest extends ContentTranslationUITestBase {
  use ImageFieldCreationTrait;
  use FileFieldCreationTrait;
  use TestFileCreationTrait;

  /**
   * {@inheritdoc}
   */
  protected $entityTypeId = 'node';

  /**
   * {@inheritdoc}
   */
  protected $bundle = 'article';

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = ['node', 'inline_translation', 'block', 'field_ui', 'image'];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected function setupBundle() {
    $this->drupalCreateContentType(['type' => 'article', 'name' => 'Article']);
    $this->createImageField('test_image', 'article', [], ['alt_field_required' => 0, 'file_directory' => 'test-upload']);
    $this->createFileField('test_file', 'node', 'article');
    // TODO: Change the autogenerated stub.
    parent::setupBundle();
  }

  /**
   * Tests the inline translation UI.
   */
  public function testTranslationUi() {
    // Ensure the help message is shown even with prefixed paths.
    $this->drupalPlaceBlock('help_block', ['region' => 'content']);

    // Display the language selector.
    $this->drupalLogin($this->administrator);
    $this->drupalGet('admin/structure/types/manage/article');
    $this->assertSession()->fieldExists('language_configuration[language_alterable]')->check();
    $this->assertSession()->buttonExists('Save content type')->press();

    // Use inline translation.
    $this->drupalGet('admin/config/regional/content-language');
    $this->assertSession()->fieldExists('settings[node][article][settings][inline_translation]')->check();
    $this->assertSession()->buttonExists('Save configuration')->press();

    $this->drupalLogin($this->translator);
    $this->drupalGet('node/add/article');
    $this->assertSession()->fieldExists('title[0][value]')->setValue('English title');
    $this->assertSession()->fieldExists('title-fr[0][value]')->setValue('French title');
    $this->assertSession()->fieldExists('title-it[0][value]')->setValue('Italian title');
    $this->assertSession()->buttonExists('Save')->press();
    $this->assertSession()->pageTextContains('Article English title has been created.');
    $this->drupalGet('node/1/edit');
    $this->assertSession()->fieldValueEquals('title[0][value]', 'English title');
    $this->assertSession()->fieldValueEquals('title-fr[0][value]', 'French title');
    $this->assertSession()->fieldValueEquals('title-it[0][value]', 'Italian title');
    $this->assertSession()->fieldExists('title[0][value]')->setValue('English title 2');
    $this->assertSession()->fieldExists('title-fr[0][value]')->setValue('French title 2');
    $this->assertSession()->fieldExists('title-it[0][value]')->setValue('Italian title 2');

    // Image field is not translatable (although the alt text is).
    $images = $this->getTestFiles('image');
    $this->getSession()->getPage()->attachFileToField('files[test_image_0]', \Drupal::service('file_system')->realpath($images[0]->uri));

    // File field is translatable.
    $files = $this->getTestFiles('text');
    $this->getSession()->getPage()->attachFileToField('files[test_file_0]', \Drupal::service('file_system')->realpath($files[0]->uri));
    $this->getSession()->getPage()->attachFileToField('files[test_file-fr_0]', \Drupal::service('file_system')->realpath($files[1]->uri));
    $this->getSession()->getPage()->attachFileToField('files[test_file-it_0]', \Drupal::service('file_system')->realpath($files[2]->uri));

    $this->assertSession()->buttonExists('Save')->press();
    $this->assertSession()->pageTextContains('Article English title 2 has been updated.');
    $this->drupalGet('node/1/edit');
    $this->assertSession()->fieldValueEquals('title[0][value]', 'English title 2');
    $this->assertSession()->fieldValueEquals('title-fr[0][value]', 'French title 2');
    $this->assertSession()->fieldValueEquals('title-it[0][value]', 'Italian title 2');
    $this->assertSession()->elementTextContains('css', '#edit-test-file-0', $files[0]->name);
    $this->assertSession()->elementTextContains('css', '#edit-test-file-fr-0', $files[1]->name);
    $this->assertSession()->elementTextContains('css', '#edit-test-file-it-0', $files[2]->name);

    $this->assertSession()->fieldExists('test_image[0][alt]')->setValue('English image alt');
    $this->assertSession()->fieldExists('test_image-fr[0][alt]')->setValue('French image alt');
    $this->assertSession()->fieldExists('test_image-it[0][alt]')->setValue('Italian image alt');
    $this->assertSession()->buttonExists('Save')->press();
    $this->assertSession()->pageTextContains('Article English title 2 has been updated.');

    $this->drupalGet('node/1/edit');
    $this->assertSession()->fieldValueEquals('test_image[0][alt]', 'English image alt');
    $this->assertSession()->fieldValueEquals('test_image-fr[0][alt]', 'French image alt');
    $this->assertSession()->fieldValueEquals('test_image-it[0][alt]', 'Italian image alt');

    // Test that inline translation works in non default languages.
    $this->drupalGet('fr/node/1/edit');
    $this->assertSession()->fieldValueEquals('test_image[0][alt]', 'French image alt');
    $this->assertSession()->fieldValueEquals('test_image-en[0][alt]', 'English image alt');
    $this->assertSession()->fieldValueEquals('test_image-it[0][alt]', 'Italian image alt');
    $this->assertSession()->fieldExists('title[0][value]')->setValue('French title 3');
    $this->assertSession()->fieldExists('title-en[0][value]')->setValue('English title 3');
    $this->assertSession()->fieldExists('title-it[0][value]')->setValue('Italian title 3');
    $this->assertSession()->buttonExists('Save')->press();
    $this->assertSession()->pageTextContains('Article French title 3 has been updated.');

    $this->drupalGet('node/1/edit');
    $this->assertSession()->fieldValueEquals('title[0][value]', 'English title 3');
    $this->assertSession()->fieldValueEquals('title-fr[0][value]', 'French title 3');
    $this->assertSession()->fieldValueEquals('title-it[0][value]', 'Italian title 3');
    $this->assertSession()->fieldValueEquals('test_image[0][alt]', 'English image alt');
    $this->assertSession()->fieldValueEquals('test_image-fr[0][alt]', 'French image alt');
    $this->assertSession()->fieldValueEquals('test_image-it[0][alt]', 'Italian image alt');

    // Make the summary part of the text body required.
    $this->drupalLogin($this->rootUser);
    $this->drupalGet("admin/structure/types/manage/article/fields/node.article.body");
    $this->assertSession()->fieldExists('settings[display_summary]')->check();
    $this->assertSession()->fieldExists('settings[required_summary]')->check();
    $this->assertSession()->buttonExists('Save settings')->press();

    $this->drupalLogin($this->administrator);
    $this->drupalGet('node/1/edit');
    $this->assertSession()->buttonExists('Save')->press();
    $this->assertSession()->pageTextContains('Summary (English) field is required.');
    $this->assertSession()->pageTextContains('Summary (French) field is required.');
    $this->assertSession()->pageTextContains('Summary (Italian) field is required.');
    $this->assertSession()->fieldExists('body[0][summary]')->setValue('English summary');
    $this->assertSession()->fieldExists('body-fr[0][summary]')->setValue('French summary');
    $this->assertSession()->fieldExists('body-it[0][summary]')->setValue('Italian summary');
    $this->assertSession()->buttonExists('Save')->press();
    $this->assertSession()->pageTextContains('Article English title 3 has been updated.');

    $this->drupalGet('node/1/edit');
    $this->assertSession()->fieldValueEquals('body[0][summary]', 'English summary');
    $this->assertSession()->fieldValueEquals('body-fr[0][summary]', 'French summary');
    $this->assertSession()->fieldValueEquals('body-it[0][summary]', 'Italian summary');

    // Test that the configuration is removed on uninstall.
    $this->assertTrue(\Drupal::entityTypeManager()->getStorage('language_content_settings')->load($this->entityTypeId . '.' . $this->bundle)->getThirdPartySetting('inline_translation', 'enabled'));
    \Drupal::service('module_installer')->uninstall(['content_translation']);
    $this->assertEmpty(\Drupal::entityTypeManager()->getStorage('language_content_settings')->load($this->entityTypeId . '.' . $this->bundle)->getThirdPartySettings('inline_translation'));
  }

  /**
   * {@inheritdoc}
   */
  protected function getTranslatorPermissions() {
    return array_merge(parent::getTranslatorPermissions(), ['administer nodes', "edit any $this->bundle content", "create $this->bundle content"]);
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditorPermissions() {
    return ['administer nodes', 'create article content'];
  }

  /**
   * {@inheritdoc}
   */
  protected function getAdministratorPermissions() {
    return array_merge(parent::getAdministratorPermissions(), ['access administration pages', 'administer content types', 'administer node fields', 'access content overview', 'bypass node access', 'administer languages', 'administer themes', 'view the administration theme']);
  }

}
